##
# File: SchemaUpdateExec.py
# Date: 11-Nov-2018  jdw
#
#  Execution wrapper  --  for schema production utilities --
#
#  Updates:
#   13-Dec-2018 jdw add Drugbank and I/HM schema options
#    7-Jan-2019 jdw overhaul
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os

from rcsb.db.utils.SchemaDefUtil import SchemaDefUtil
from rcsb.utils.config.ConfigUtil import ConfigUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s')
logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser()
    #
    defaultConfigName = 'site_info'
    #
    parser.add_argument("--update_chem_comp_ref", default=False, action='store_true', help="Update schema for Chemical Component reference definitions")
    parser.add_argument("--update_chem_comp_core_ref", default=False, action='store_true', help="Update core schema for Chemical Component reference definitions")
    parser.add_argument("--update_bird_chem_comp_ref", default=False, action='store_true', help="Update schema for Bird Chemical Component reference definitions")
    parser.add_argument("--update_bird_chem_comp_core_ref", default=False, action='store_true', help="Update core schema for Bird Chemical Component reference definitions")

    parser.add_argument("--update_bird_ref", default=False, action='store_true', help="Update schema for Bird reference definitions")
    parser.add_argument("--update_bird_family_ref", default=False, action='store_true', help="Update schema for Bird Family reference definitions")

    parser.add_argument("--update_pdbx", default=False, action='store_true', help="Update schema for PDBx entry data")
    parser.add_argument("--update_pdbx_core", default=False, action='store_true', help="Update schema for PDBx core entry/entity data")
    #
    parser.add_argument("--update_repository_holdings", default=False, action='store_true', help="Update schema for repository holdings")
    parser.add_argument("--update_entity_sequence_clusters", default=False, action='store_true', help="Update schema for entity sequence clusters")
    parser.add_argument("--update_data_exchange", default=False, action='store_true', help="Update schema for data exchange status")
    parser.add_argument("--update_ihm_dev", default=False, action='store_true', help="Update schema for I/HM dev entry data")
    parser.add_argument("--update_drugbank_core", default=False, action='store_true', help="Update DrugBank schema")
    #
    parser.add_argument("--update_config_all", default=False, action='store_true', help="Update using configuration settings (e.g. SCHEMA_NAMES_ALL)")
    parser.add_argument("--update_config_deployed", default=False, action='store_true', help="Update using configuration settings (e.g. SCHEMA_NAMES_DEPLOYED)")
    parser.add_argument("--update_config_test", default=False, action='store_true', help="Update using configuration settings (e.g. SCHEMA_NAMES_TEST)")
    #
    parser.add_argument("--config_path", default=None, help="Path to configuration options file")
    parser.add_argument("--config_name", default=defaultConfigName, help="Configuration section name")
    #
    parser.add_argument("--schema_dirpath", default=None, help="Output schema directory path (overrides configuration settings)")
    parser.add_argument("--schema_types", default=None, help="Schema encoding (rcsb|json|bson) (comma separated)")
    parser.add_argument("--schema_levels", default=None, help="Schema validation level (full|min) (comma separated)")
    #
    parser.add_argument("--debug", default=False, action='store_true', help="Turn on verbose logging")
    parser.add_argument("--mock", default=False, action='store_true', help="Use MOCK repository configuration for dependencies and testing")
    # parser.add_argument("--working_path", default=None, help="Working/alternative path for temporary and schema files")
    args = parser.parse_args()
    #
    debugFlag = args.debug
    if debugFlag:
        logger.setLevel(logging.DEBUG)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #                                       Configuration Details
    configPath = args.config_path
    configName = args.config_name
    schemaDirPath = args.schema_dirpath
    #
    schemaTypes = args.schema_types.split(',') if args.schema_types else []
    schemaLevels = args.schema_levels.split(',') if args.schema_levels else []
    dataTypingList = ['ANY', 'SQL']

    if not configPath:
        configPath = os.getenv('DBLOAD_CONFIG_PATH', None)
    try:
        if os.access(configPath, os.R_OK):
            os.environ['DBLOAD_CONFIG_PATH'] = configPath
            logger.info("Using configuation path %s (%s)" % (configPath, configName))
        else:
            logger.error("Missing or access issue with config file %r" % configPath)
            exit(1)
        mockTopPath = os.path.join(TOPDIR, 'rcsb', 'mock-data') if args.mock else None
        cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=defaultConfigName, mockTopPath=mockTopPath)
        if configName != defaultConfigName:
            cfgOb.replaceSectionName(defaultConfigName, configName)
    except Exception as e:
        logger.error("Missing or access issue with config file %r with %s" % (configPath, str(e)))
        exit(1)
    #
    schemaNameList = []
    if args.update_chem_comp_ref:
        schemaNameList.append('chem_comp')

    if args.update_bird_chem_comp_ref:
        schemaNameList.append('bird_chem_comp')

    if args.update_chem_comp_core_ref:
        schemaNameList.append('chem_comp_core')

    if args.update_bird_chem_comp_core_ref:
        schemaNameList.append('bird_chem_comp_core')

    if args.update_bird_ref:
        schemaNameList.append('bird')

    if args.update_bird_family_ref:
        schemaNameList.append('bird_family')

    if args.update_pdbx:
        schemaNameList.append('pdbx')

    if args.update_pdbx_core:
        schemaNameList.append('pdbx_core')

    if args.update_repository_holdings:
        schemaNameList.append('repository_holdings')

    if args.update_entity_sequence_clusters:
        schemaNameList.append('entity_sequence_clusters')

    if args.update_data_exchange:
        schemaNameList.append('data_exchange')

    if args.update_ihm_dev:
        schemaNameList.append('ihm_dev')

    if args.update_drugbank_core:
        schemaNameList.append('drugbank_core')

    if args.update_config_deployed:
        schemaNameList = cfgOb.getList('SCHEMA_NAMES_DEPLOYED', sectionName='schema_catalog_info')
        dataTypingList = cfgOb.getList('DATATYPING_DEPLOYED', sectionName='schema_catalog_info')
        schemaLevels = cfgOb.getList('SCHEMA_LEVELS_DEPLOYED', sectionName='schema_catalog_info')
        schemaTypes = cfgOb.getList('SCHEMA_TYPES_DEPLOYED', sectionName='schema_catalog_info')

    if args.update_config_all:
        schemaNameList = cfgOb.getList('SCHEMA_NAMES_ALL', sectionName='schema_catalog_info')
        dataTypingList = cfgOb.getList('DATATYPING_ALL', sectionName='schema_catalog_info')
        schemaLevels = cfgOb.getList('SCHEMA_LEVELS_ALL', sectionName='schema_catalog_info')
        schemaTypes = cfgOb.getList('SCHEMA_TYPES_ALL', sectionName='schema_catalog_info')

    if args.update_config_test:
        schemaNameList = cfgOb.getList('SCHEMA_NAMES_TEST', sectionName='schema_catalog_info')
        dataTypingList = cfgOb.getList('DATATYPING_TEST', sectionName='schema_catalog_info')
        schemaLevels = cfgOb.getList('SCHEMA_LEVELS_TEST', sectionName='schema_catalog_info')
        schemaTypes = cfgOb.getList('SCHEMA_TYPES_TEST', sectionName='schema_catalog_info')
    #
    scnD = cfgOb.get('schema_collection_names', sectionName='document_helper')
    #
    schemaNameList = list(set(schemaNameList))
    logger.debug("Collections %s" % (list(scnD.items())))
    logger.debug("schemaNameList %s" % schemaNameList)

    sdu = SchemaDefUtil(cfgOb)
    for schemaName in schemaNameList:
        for schemaType in schemaTypes:
            if schemaType == 'rcsb':
                for dataTyping in dataTypingList:
                    logger.info("Creating schema definition for content type %s data typing %s" % (schemaName, dataTyping))
                    sdu.makeSchemaDef(schemaName, dataTyping=dataTyping, saveSchema=True, altDirPath=schemaDirPath)
            else:
                if schemaName in scnD:
                    for d in scnD[schemaName]:
                        collectionName = d['NAME']
                        for schemaLevel in schemaLevels:
                            logger.info("Creating %r schema for content type %s collection %s" % (schemaType, schemaName, collectionName))
                            sdu.makeSchema(schemaName, collectionName, schemaType=schemaType, level=schemaLevel, saveSchema=True, altDirPath=schemaDirPath)


if __name__ == '__main__':
    main()



##
# File:    SchemaDefUtil.py
# Author:  J. Westbrook
# Date:    31-Mar-2018
# Version: 0.001
#
# Updates:
#   9-Apr-2018 jdw update to provide indices and include remaining schema
#  18-Jun-2018 jdw update to new schema generation protocol
#  22-Jun-2018 jdw change collection attribute specification to dot notation.
#  14-Aug-2018 jdw generalize the primaryIndex to the list of attributes returned by getDocumentKeyAttributeNames()
#  21-Aug-2018 jdw use getHelper() from the configuration class.
#   7-Sep-2018 jdw add method to return a stored JSON schema getJsonSchema()
#  11-Nov-2018 jdw add support for chem_comp_core and bird_chem_comp_core schemas
#   3-Dec-2018 jdw generalize the delivery of document indices
#   4-Jan-2019 jdw add prefix path options to api and json/bson schema paths
#                  add method getJsonSchemaLocator()
#   8-Jan-2019 jdw standardize argument names -
#   6-Feb-2019 jdw add option to merge content types to getPathList()
#
##
"""
 A collection of schema and repo path convenience methods.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import os

from rcsb.db.define.SchemaDefAccess import SchemaDefAccess
from rcsb.db.define.SchemaDefBuild import SchemaDefBuild
from rcsb.db.utils.HashableDict import HashableDict
from rcsb.db.utils.RepoPathUtil import RepoPathUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class SchemaDefUtil(object):
    """ A collection of schema and repository convenience methods.
    """

    def __init__(self, cfgOb=None, numProc=1, fileLimit=None, workPath=None, **kwargs):
        self.__cfgOb = cfgOb
        self.__fileLimit = fileLimit
        self.__numProc = numProc
        self.__workPath = workPath

    def getSchemaOptions(self, schemaLevel, extraOpts=None):
        opts = extraOpts + '|' if extraOpts else ""
        if schemaLevel == 'full':
            return opts + "mandatoryKeys|mandatoryAttributes|bounds|enums"
        elif schemaLevel in ['min', 'minimum']:
            return opts + "mandatoryKeys|enums"
        else:
            return opts

    def getLocatorObjList(self, contentType, inputPathList=None, mergeContentTypes=None):
        """Convenience method to get the data path list for the input repository content type.

        Args:
            contentType (str): Repository content type (e.g. pdbx, chem_comp, bird, ...)
            inputPathList (list, optional): path list that will be returned if provided.
            mergeContentTypes (list, optional): repository content types to combined with the
                              primary content type.

        Returns:
            Obj list: data file paths or tuple of file paths


        """
        inputPathList = inputPathList if inputPathList else []
        rpU = RepoPathUtil(self.__cfgOb, numProc=self.__numProc, fileLimit=self.__fileLimit, workPath=self.__workPath)
        locatorList = rpU.getLocatorList(contentType, inputPathList=inputPathList)
        #
        if mergeContentTypes and 'vrpt' in mergeContentTypes and contentType in ['pdbx', 'pdbx_core']:
            dictMapPath = self.__cfgOb.getPath('VRPT_DICT_MAPPING_LOCATOR', sectionName=self.__cfgOb.getDefaultSectionName())
            locObjL = []
            for locator in locatorList:
                if isinstance(locator, str):
                    kwD = HashableDict({})
                    oL = [HashableDict({'locator': locator, 'format': 'mmcif', 'kwargs': kwD})]
                    for mergeContentType in mergeContentTypes:
                        pth, fn = os.path.split(locator)
                        idCode = fn[:4] if fn and len(fn) >= 8 else None
                        mergeLocator = rpU.getLocator(mergeContentType, idCode) if idCode else None
                        if mergeLocator:
                            kwD = HashableDict({'dictMapPath': dictMapPath})
                            oL.append(HashableDict({'locator': mergeLocator, 'format': 'vrpt-xml-to-cif', 'kwargs': kwD}))
                    lObj = tuple(oL)
                else:
                    logger.error("Unexpected output locator type %r" % locator)
                    lObj = locator
                locObjL.append(lObj)
            #
            locatorList = locObjL

        # -
        return locatorList

    def getSchemaInfo(self, contentType, dataTyping='ANY', altDirPath=None):
        """Convenience method to return essential schema details for the input repository content type.

        Args:
            contentType (str): Repository content type (e.g. pdbx, bird, chem_comp, ...)
            dataTyping (str, optional): Application name for the target schema (e.g. ANY, SQL, ...)
            altDirPath (str, optional): alternative directory path for schema definition file

        Returns:
            tuple: SchemaDefAccess(object), target database name, target collection name list, primary index attribute list


        """
        sd = None
        dbName = None
        collectionNameList = []
        docIndexD = {}
        try:
            mU = MarshalUtil(workPath=self.__workPath)
            schemaLocator = self.getSchemaDefLocator(contentType, dataTyping=dataTyping, altDirPath=altDirPath)
            logger.debug("ContentType %r dataTyping %r altDirPath %r schemaLocator %r " % (contentType, dataTyping, altDirPath, schemaLocator))
            schemaDef = mU.doImport(schemaLocator, format="json")
            if schemaDef:
                logger.debug("Using cached schema definition for %s application %s" % (contentType, dataTyping))
                sd = SchemaDefAccess(schemaDef)
                if sd:
                    dbName = sd.getDatabaseName()
                    collectionInfoList = sd.getCollectionInfo()
                    logger.debug("Schema %s database name %s collections %r" % (contentType, dbName, collectionInfoList))
                    for cd in collectionInfoList:
                        collectionName = cd['NAME']
                        collectionNameList.append(collectionName)
                        docIndexD[collectionName] = sd.getDocumentIndices(collectionName)

        except Exception as e:
            logger.exception("Retreiving schema %s for %s failing with %s" % (contentType, dataTyping, str(e)))

        return sd, dbName, collectionNameList, docIndexD

    def getSchemaDefLocator(self, contentType, dataTyping='ANY', altDirPath=None):
        """Return schema definition path for the input content type and application.

           Defines schema definition naming convention -

           Uses configuration details for directory path/locator details.

           site_info:
            'SCHEMA_DEF_LOCATOR_PATH': <locator prefix for schema file

           Args:
            contentType (str): Repository content type (e.g. pdbx, bird, chem_comp, ...)
            dataTyping (str, optional): Application name for the target schema (e.g. ANY, SQL, ...)
            altDirPath (str, optional): alternative directory path for schema definition file

            Returns:

             str: schema definition file locator

        """
        schemaLocator = None
        try:
            prefixName = 'SCHEMA_DEF_LOCATOR_PATH'
            pth = self.__cfgOb.getPath(prefixName, sectionName=self.__cfgOb.getDefaultSectionName())
            fn = 'schema_def-%s-%s.json' % (contentType, dataTyping.upper())
            #
            schemaLocator = os.path.join(altDirPath, fn) if altDirPath else os.path.join(pth, fn)
        except Exception as e:
            logger.exception("Retreiving schema definition path %s for %s failing with %s" % (contentType, dataTyping, str(e)))
        return schemaLocator

    def getJsonSchemaLocator(self, collectionName, schemaType='BSON', level='full', altDirPath=None):
        """Return JSON schema path for the input collection data type convention and level.

           Defines the JSON/BSON schema naming convention -

           site_info:
            'JSON_SCHEMA_LOCATOR_PATH': <locator prefix for schema file>

           Args:
            collectionName (str): Collection name in document store
            schemaType (str, optional): data type convention (BSON|JSON)
            level (str, optional): Completeness of the schema (e.g. min or full)
            altDirPath (str, optional): alternative directory path for json schema file

            Returns:

            str: schema file locator

        """
        schemaLocator = None
        try:
            prefixName = 'JSON_SCHEMA_LOCATOR_PATH'
            sdType = None
            sLevel = None
            schemaLocator = None
            if schemaType.upper() in ['JSON', 'BSON']:
                sdType = schemaType.lower()
            if level.lower() in ['min', 'minimun']:
                sLevel = 'min'
            elif level.lower() in ['full']:
                sLevel = level.lower()
            #
            if sdType and sLevel:
                pth = self.__cfgOb.getPath(prefixName, sectionName=self.__cfgOb.getDefaultSectionName())
                fn = "%s-schema-%s-%s.json" % (sdType, sLevel, collectionName)
                schemaLocator = os.path.join(altDirPath, fn) if altDirPath else os.path.join(pth, fn)
            else:
                logger.error("Unsupported schema options:  %s level %r type %r" % (collectionName, level, schemaType))
                schemaLocator = None
        except Exception as e:
            logger.debug("Retreiving JSON schema definition for %s type %s failing with %s" % (collectionName, schemaType, str(e)))
        #
        return schemaLocator

    def getJsonSchema(self, collectionName, schemaType='BSON', level='full', altDirPath=None):
        """Return JSON schema (w/ BSON types) object for the input collection and level.and

           Currenting using configuration information -



        Args:
            collectionName (str): Collection name in document store
            schemaType (str, optional): data type convention (BSON|JSON)
            level (str, optional): Completeness of the schema (e.g. min or full)
            altDirPath (str, optional): alternative directory path for json schema file

        Returns:
            dict: Schema object

        """
        sObj = None
        schemaLocator = self.getJsonSchemaLocator(collectionName, schemaType=schemaType, level=level, altDirPath=altDirPath)
        mU = MarshalUtil(workPath=self.__workPath)
        if schemaLocator and mU.exists(schemaLocator):
            mU = MarshalUtil(workPath=self.__workPath)
            sObj = mU.doImport(schemaLocator, format="json")
        else:
            logger.debug("Failed to read schema for %s %r" % (collectionName, level))
        return sObj

    def makeSchema(self, schemaName, collectionName, schemaType='BSON', level='full', saveSchema=False, altDirPath=None, extraOpts=None):
        try:
            smb = SchemaDefBuild(schemaName, self.__cfgOb)
            #
            cD = None
            stU = schemaType.upper()
            cD = smb.build(collectionName, dataTyping=stU, schemaType=stU, enforceOpts=self.getSchemaOptions(level, extraOpts=extraOpts))
            if cD and saveSchema:
                schemaLocator = self.getJsonSchemaLocator(collectionName, schemaType=schemaType, level=level, altDirPath=altDirPath)
                mU = MarshalUtil(workPath=self.__workPath)
                mU.doExport(schemaLocator, cD, format="json", indent=3)
        except Exception as e:
            logger.exception("Building schema %s collection %s failing with %s" % (schemaName, collectionName, str(e)))
        return cD

    def makeSchemaDef(self, schemaName, dataTyping='ANY', saveSchema=False, altDirPath=None):
        schemaDef = None
        try:
            smb = SchemaDefBuild(schemaName, self.__cfgOb)
            schemaDef = smb.build(dataTyping=dataTyping, schemaType='rcsb')
            if schemaDef and saveSchema:
                schemaLocator = self.getSchemaDefLocator(schemaName, dataTyping=dataTyping, altDirPath=altDirPath)
                mU = MarshalUtil(workPath=self.__workPath)
                mU.doExport(schemaLocator, schemaDef, format="json", indent=3)
        except Exception as e:
            logger.exception("Building schema %s failing with %s" % (schemaName, str(e)))
        return schemaDef



#
# from six.moves.html_parser import HTMLParser

try:
    # Python 2.6-2.7
    from HTMLParser import unescape
except ImportError:
    # Python 3.6+
    from html import unescape


def unescapeXmlCharRef(i_str):
    """
    Convert html character entities into unicode.
    """
    o_str = unescape(i_str)

    return o_str


if __name__ == '__main__':
    print("START")
    print("%r" % unescapeXmlCharRef('&lt;b&gt;'))
    print("%r" % unescapeXmlCharRef('Here is a &quot;').encode('utf-8'))
    print("%r" % unescapeXmlCharRef('Here is a &Phi;').encode('utf-8'))
    print("%range" % unescapeXmlCharRef('Here is a &Psi;'))
    print("%r" % unescapeXmlCharRef('Here is a &alpha;'))
    print("%r" % unescapeXmlCharRef('Here is a &#xa3;'))

    print("%r" % unescapeXmlCharRef('Here is a &#8453;'))
    print("%r" % unescapeXmlCharRef('Here is a &#9734;'))
    print("%r" % unescapeXmlCharRef('Here is a &#120171;'))
    for ichar in range(1, 8000):
        myStr = "decimal %6d char &#%d;" % (ichar, ichar)
        print("%r" % unescapeXmlCharRef(myStr))



##
# File: ChemRefEtlWorker.py
# Date: 2-Jul-2018  jdw
#
# ETL utilities for processing chemical reference data and related data integration.
#
# Updates:
#  9-Dec-2018  jdw add validation methods
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.db.mongo.ChemRefExtractor import ChemRefExtractor
from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.processors.DataExchangeStatus import DataExchangeStatus
from rcsb.db.utils.SchemaDefUtil import SchemaDefUtil
from rcsb.utils.chemref.ChemRefDataPrep import ChemRefDataPrep

logger = logging.getLogger(__name__)


class ChemRefEtlWorker(object):
    """ Prepare and load chemical reference data collections.
    """

    def __init__(self, cfgOb, workPath=None, numProc=2, chunkSize=10, readBackCheck=False, documentLimit=None, mockTopPath=None, verbose=False):
        self.__cfgOb = cfgOb
        self.__workPath = workPath
        self.__readBackCheck = readBackCheck
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        self.__documentLimit = documentLimit
        self.__mockTopPath = mockTopPath
        #
        self.__resourceName = "MONGO_DB"
        self.__filterType = "assign-dates"
        self.__verbose = verbose
        self.__statusList = []
        self.__schU = SchemaDefUtil(cfgOb=self.__cfgOb, numProc=self.__numProc, fileLimit=self.__documentLimit, workPath=self.__workPath)
        #

    def __updateStatus(self, updateId, databaseName, collectionName, status, startTimestamp):
        try:
            sFlag = 'Y' if status else 'N'
            desp = DataExchangeStatus()
            desp.setStartTime(tS=startTimestamp)
            desp.setObject(databaseName, collectionName)
            desp.setStatus(updateId=updateId, successFlag=sFlag)
            desp.setEndTime()
            self.__statusList.append(desp.getStatus())
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def load(self, updateId, extResource, loadType='full'):
        """ Load chemical reference integrated data for the input external resource-

        Relevant configuration options:

        site_info:
            DRUGBANK_MAPPING_LOCATOR: DrugBank/drugbank_pdb_mapping.json
            DRUGBANK_DATA_LOCATOR: DrugBank/full_database.xml.gz

        """
        try:
            self.__statusList = []
            desp = DataExchangeStatus()
            statusStartTimestamp = desp.setStartTime()
            #
            crdp = ChemRefDataPrep(self.__cfgOb)
            crExt = ChemRefExtractor(self.__cfgOb)

            idD = crExt.getChemCompAccesionMapping(extResource)
            dList = crdp.getDocuments(extResource, idD)
            #
            logger.info("Resource %r mapped document length %d" % (extResource, len(dList)))
            logger.debug("Objects %r" % dList[:2])

            schemaName = 'drugbank_core'
            sD, databaseName, collectionList, _ = self.__schU.getSchemaInfo(schemaName)
            collectionName = collectionList[0] if len(collectionList) > 0 else 'unassigned'
            indexL = sD.getDocumentIndex(collectionName, 'primary')
            logger.info("Database %r collection %r index attributes %r " % (databaseName, collectionName, indexL))
            #
            collectionVersion = sD.getCollectionVersion(collectionName)
            addValues = {'_schema_version': collectionVersion}
            #
            dl = DocumentLoader(self.__cfgOb, self.__resourceName, numProc=self.__numProc, chunkSize=self.__chunkSize,
                                documentLimit=self.__documentLimit, verbose=self.__verbose, readBackCheck=self.__readBackCheck)
            # JDW --
            # databaseName = self.__cfgOb.get('DATABASE_NAME', sectionName=sectionName) + '_' + self.__cfgOb.get('DATABASE_VERSION_STRING', sectionName=sectionName)
            # collectionVersion = self.__cfgOb.get('COLLECTION_VERSION_STRING', sectionName=sectionName)
            # collectionName = self.__cfgOb.get('COLLECTION_DRUGBANK_CORE', sectionName=sectionName) + '_' + collectionVersion
            #
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList,
                         indexAttributeList=indexL, keyNames=None, addValues=addValues)

            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def getLoadStatus(self):
        return self.__statusList



##
# File:    RepoPathUtil.py
# Author:  J. Westbrook
# Date:    21-Mar-2018
#
# Updates:
#   22-Mar-2018  jdw add support for all repositories -
#   26-Mar-2018  jdw internalize the use of externally provided configuration object -
#   27-Mar-2018  jdw add path to support mock repositories for testing.
#   23-May-2018  jdw add getRepoPathList() convenience method
#   18-Jun-2018  jdw move mock support to the configuration module
#   12-Jul-2018  jdw correct config for PDBX_REPO_PATH
#   13-Aug-2018  jdw add support for gz compressed entry files
#   24-Oct-2018  jdw update for new configuration organization
#   28-Nov-2018  jdw add mergeBirdRefData()
#   13-Dec-2018  jdw add preliminary I/HM repository support
#    5-Feb-2019  jdw add just method naming conventions, add getLocator() method,
#                    consolidate deliver of path configuration details in __getRepoTopPath().
#   14-Mar-2019  jdw add VRPT_REPO_PATH_ENV as an override for the validation report repo path.
#
#
##
"""
 Utilites for scanning and accessing data in common repository file systems.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import time

from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

try:
    from os import walk
except ImportError:
    from scandir import walk


logger = logging.getLogger(__name__)


class RepoPathUtil(object):

    def __init__(self, cfgOb, cfgSectionName='site_info', numProc=8, fileLimit=None, workPath=None, verbose=False):
        self.__fileLimit = fileLimit
        self.__numProc = numProc
        self.__verbose = verbose
        self.__cfgOb = cfgOb
        self.__cfgSectionName = cfgSectionName
        self.__workPath = workPath if workPath else '.'
        #
        self.__mpFormat = '[%(levelname)s] %(asctime)s %(processName)s-%(module)s.%(funcName)s: %(message)s'

    def getLocatorList(self, contentType, inputPathList=None):
        """ Convenience method to return repository path list by content type:
        """
        outputPathList = []
        inputPathList = inputPathList if inputPathList else []
        try:
            if contentType == "bird":
                outputPathList = inputPathList if inputPathList else self.getBirdPathList()
            elif contentType == "bird_family":
                outputPathList = inputPathList if inputPathList else self.getBirdFamilyPathList()
            elif contentType in ['chem_comp', 'chem_comp_core']:
                outputPathList = inputPathList if inputPathList else self.getChemCompPathList()
            elif contentType in ['bird_chem_comp']:
                outputPathList = inputPathList if inputPathList else self.getBirdChemCompPathList()
            elif contentType in ['pdbx', 'pdbx_core']:
                outputPathList = inputPathList if inputPathList else self.getEntryPathList()
            elif contentType in ['bird_consolidated', 'bird_chem_comp_core']:
                outputPathList = inputPathList if inputPathList else self.mergeBirdRefData()
            elif contentType in ['ihm_dev', 'ihm_dev_core', 'ihm_dev_full']:
                outputPathList = inputPathList if inputPathList else self.getIhmDevPathList()
            elif contentType in ['pdb_distro', 'da_internal', 'status_history']:
                outputPathList = inputPathList if inputPathList else []
            else:
                logger.warning("Unsupported contentType %s" % contentType)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        if self.__fileLimit:
            outputPathList = outputPathList[:self.__fileLimit]

        return outputPathList

    def getLocator(self, contentType, idCode, version='v1-0'):
        """ Convenience method to return repository path for a content type and cardinal identifier.
        """
        pth = None
        try:
            idCodel = idCode.lower()
            if contentType == "bird":
                pth = os.path.join(self.__getRepoTopPath(contentType), idCode[-1], idCode + '.cif')
            elif contentType == "bird_family":
                pth = os.path.join(self.__getRepoTopPath(contentType), idCode[-1], idCode + '.cif')
            elif contentType in ['chem_comp', 'chem_comp_core']:
                pth = os.path.join(self.__getRepoTopPath(contentType), idCode[0], idCode, idCode + '.cif')
            elif contentType in ['bird_chem_comp']:
                pth = os.path.join(self.__getRepoTopPath(contentType), idCode[-1], idCode + '.cif')
            elif contentType in ['pdbx', 'pdbx_core']:
                pth = os.path.join(self.__getRepoTopPath(contentType), idCodel[1:3], idCodel, idCodel + '.cif.gz')
            elif contentType in ['bird_consolidated', 'bird_chem_comp_core']:
                pth = os.path.join(self.__getRepoTopPath(contentType), idCode + '.cif')
            elif contentType in ['ihm_dev', 'ihm_dev_core', 'ihm_dev_full']:
                pth = os.path.join(self.__getRepoTopPath(contentType), idCode, idCode + '_model_%s.cif.gz' % version)
            elif contentType in ['pdb_distro', 'da_internal', 'status_history']:
                pass
            elif contentType in ['vrpt']:
                pth = os.path.join(self.__getRepoTopPath(contentType), idCodel[1:3], idCodel, idCodel + '_validation.xml.gz')
            else:
                logger.warning("Unsupported contentType %s" % contentType)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return pth

    def __getRepoTopPath(self, contentType):
        """ Convenience method to return repository top path from configuration data.
        """
        pth = None
        try:
            if contentType == "bird":
                pth = self.__cfgOb.getPath('BIRD_REPO_PATH', sectionName=self.__cfgSectionName)
            elif contentType == "bird_family":
                pth = self.__cfgOb.getPath('BIRD_FAMILY_REPO_PATH', sectionName=self.__cfgSectionName)
            elif contentType in ['chem_comp', 'chem_comp_core']:
                pth = self.__cfgOb.getPath('CHEM_COMP_REPO_PATH', sectionName=self.__cfgSectionName)
            elif contentType in ['bird_chem_comp']:
                pth = self.__cfgOb.getPath('BIRD_CHEM_COMP_REPO_PATH', sectionName=self.__cfgSectionName)
            elif contentType in ['pdbx', 'pdbx_core']:
                pth = self.__cfgOb.getPath('PDBX_REPO_PATH', sectionName=self.__cfgSectionName)
            elif contentType in ['bird_consolidated', 'bird_chem_comp_core']:
                pth = self.__workPath
            elif contentType in ['ihm_dev', 'ihm_dev_core', 'ihm_dev_full']:
                pth = self.__cfgOb.getPath('IHM_DEV_REPO_PATH', sectionName=self.__cfgSectionName)
            elif contentType in ['pdb_distro', 'da_internal', 'status_history']:
                pass
            elif contentType in ['vrpt']:
                pth = self.__cfgOb.getEnvValue('VRPT_REPO_PATH_ENV', sectionName=self.__cfgSectionName, default=None)
                if pth is None:
                    pth = self.__cfgOb.getPath('VRPT_REPO_PATH', sectionName=self.__cfgSectionName)
                else:
                    logger.debug("Using validation report path from environment assignment %s" % pth)
            else:
                logger.warning("Unsupported contentType %s" % contentType)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return pth

    def _chemCompPathWorker(self, dataList, procName, optionsD, workingDir):
        """ Return the list of chemical component definition file paths in the current repository.
        """
        topRepoPath = optionsD['topRepoPath']
        pathList = []
        for subdir in dataList:
            dd = os.path.join(topRepoPath, subdir)
            for root, dirs, files in walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.endswith(".cif") and len(name) <= 7:
                        pathList.append(os.path.join(root, name))
        return dataList, pathList, []

    def getChemCompPathList(self):
        return self.__getChemCompPathList(self.__getRepoTopPath('chem_comp'), numProc=self.__numProc)

    def __getChemCompPathList(self, topRepoPath, numProc=8):
        """Get the path list for the chemical component definition repository
        """
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting at %s" % ts)
        startTime = time.time()
        pathList = []
        try:
            dataS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
            dataList = [a for a in dataS]
            optD = {}
            optD['topRepoPath'] = topRepoPath
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_chemCompPathWorker")
            ok, failList, retLists, diagList = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Path list length %d  in %.4f seconds\n" % (len(pathList), endTime0 - startTime))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return self.__applyFileLimit(pathList)

    def _entryPathWorker(self, dataList, procName, optionsD, workingDir):
        """ Return the list of entry file paths in the current repository.
        """
        topRepoPath = optionsD['topRepoPath']
        pathList = []
        for subdir in dataList:
            dd = os.path.join(topRepoPath, subdir)
            for root, dirs, files in walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if ((name.endswith(".cif.gz") and len(name) == 11) or (name.endswith(".cif") and len(name) == 8)):
                        pathList.append(os.path.join(root, name))
        return dataList, pathList, []

    def getEntryPathList(self):
        return self.__getEntryPathList(self.__getRepoTopPath('pdbx'), numProc=self.__numProc)

    def __getEntryPathList(self, topRepoPath, numProc=8):
        """Get the path list for structure entries in the input repository
        """
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting at %s" % ts)
        startTime = time.time()
        pathList = []
        try:
            dataList = []
            anL = 'abcdefghijklmnopqrstuvwxyz0123456789'
            for a1 in anL:
                for a2 in anL:
                    hc = a1 + a2
                    dataList.append(hc)
                    hc = a2 + a1
                    dataList.append(hc)
            dataList = list(set(dataList))
            #
            optD = {}
            optD['topRepoPath'] = topRepoPath
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_entryPathWorker")
            ok, failList, retLists, diagList = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Path list length %d  in %.4f seconds\n" % (len(pathList), endTime0 - startTime))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return self.__applyFileLimit(pathList)

    def getBirdPathList(self):
        return self.__getBirdPathList(self.__getRepoTopPath('bird'))

    def __getBirdPathList(self, topRepoPath):
        """ Return the list of definition file paths in the current repository.

            List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, dirs, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PRD_") and name.endswith(".cif") and len(name) <= 14:
                        pth = os.path.join(root, name)
                        sd[int(name[4:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return self.__applyFileLimit(pathList)

    def getBirdFamilyPathList(self):
        return self.__getBirdFamilyPathList(self.__getRepoTopPath('bird_family'))

    def __getBirdFamilyPathList(self, topRepoPath):
        """ Return the list of definition file paths in the current repository.

            List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, dirs, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("FAM_") and name.endswith(".cif") and len(name) <= 14:
                        pth = os.path.join(root, name)
                        sd[int(name[4:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return self.__applyFileLimit(pathList)

    def getBirdChemCompPathList(self):
        return self.__getBirdChemCompPathList(self.__getRepoTopPath('bird_chem_comp'))

    def __getBirdChemCompPathList(self, topRepoPath):
        """ Return the list of definition file paths in the current repository.

            List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, dirs, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PRDCC_") and name.endswith(".cif") and len(name) <= 16:
                        pth = os.path.join(root, name)
                        sd[int(name[6:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return self.__applyFileLimit(pathList)

    def __applyFileLimit(self, pathList):
        logger.debug("Length of file path list %d (limit %r)" % (len(pathList), self.__fileLimit))
        if self.__fileLimit:
            return pathList[:self.__fileLimit]
        else:
            return pathList

    def __buildFamilyIndex(self):
        """ Using information from the PRD family definition:
            #
            loop_
            _pdbx_reference_molecule_list.family_prd_id
            _pdbx_reference_molecule_list.prd_id
                FAM_000010 PRD_000041
                FAM_000010 PRD_000042
                FAM_000010 PRD_000043
                FAM_000010 PRD_000044
                FAM_000010 PRD_000048
                FAM_000010 PRD_000049
                FAM_000010 PRD_000051
            #
        """
        prdD = {}
        try:
            mU = MarshalUtil(workPath=self.__workPath)
            pthL = self.getLocatorList('bird_family')
            for pth in pthL:
                containerL = mU.doImport(pth, format="mmcif")
                for container in containerL:
                    catName = "pdbx_reference_molecule_list"
                    if container.exists(catName):
                        catObj = container.getObj(catName)
                        for ii in range(catObj.getRowCount()):
                            familyPrdId = catObj.getValue(attributeName='family_prd_id', rowIndex=ii)
                            prdId = catObj.getValue(attributeName='prd_id', rowIndex=ii)
                            if prdId in prdD:
                                logger.debug("duplicate prdId in family index %s %s " % (prdId, familyPrdId))
                            prdD[prdId] = {'familyPrdId': familyPrdId, 'c': container}
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return prdD
    #

    def mergeBirdRefData(self):
        """ Consolidate all of the bird reference data in a single container.

            Return a path list for the consolidated data files -

        """
        outPathList = []
        try:
            mU = MarshalUtil(workPath=self.__workPath)
            birdPathList = self.getLocatorList('bird')
            birdPathD = {}
            for birdPath in birdPathList:
                _, fn = os.path.split(birdPath)
                prdId, _ = os.path.splitext(fn)
                birdPathD[prdId] = birdPath
            #
            logger.debug("BIRD data length %d" % len(birdPathD))
            logger.debug("BIRD keys %r" % list(birdPathD.keys()))
            birdCcPathList = self.getLocatorList('bird_chem_comp')
            birdCcPathD = {}
            for birdCcPath in birdCcPathList:
                _, fn = os.path.split(birdCcPath)
                prdCcId, _ = os.path.splitext(fn)
                prdId = 'PRD_' + prdCcId[6:]
                birdCcPathD[prdId] = birdCcPath
            #
            logger.debug("BIRD CC data length %d" % len(birdCcPathD))
            logger.debug("BIRD CC keys %r" % list(birdCcPathD.keys()))
            fD = self.__buildFamilyIndex()
            logger.debug("Family index length %d" % len(fD))
            logger.debug("Family index keys %r" % list(fD.keys()))
            #
            #
            for prdId in birdPathD:
                fp = os.path.join(self.__workPath, prdId + '.cif')
                logger.debug("Export path is %r" % fp)
                #
                if prdId in birdPathD:
                    pth2 = birdPathD[prdId]
                    c2L = mU.doImport(pth2, format="mmcif")
                    cFull = c2L[0]
                    logger.debug("Got cBird %r" % cFull.getName())
                #
                ccBird = None
                if prdId in birdCcPathD:
                    pth1 = birdCcPathD[prdId]
                    c1L = mU.doImport(pth1, format="mmcif")
                    ccBird = c1L[0]
                    logger.debug("Got cFull %r" % ccBird.getName())
                    #
                cFam = None
                if prdId in fD:
                    cFam = fD[prdId]['c']
                    logger.debug("Got cFam %r" % cFam.getName())
                #
                if cFam:
                    for catName in cFam.getObjNameList():
                        cFull.append(cFam.getObj(catName))
                #
                if ccBird:
                    for catName in ccBird.getObjNameList():
                        cFull.append(ccBird.getObj(catName))
                #
                mU.doExport(fp, [cFull], format='mmcif')
                outPathList.append(fp)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return outPathList
        #

    def __exportConfig(self, container):
        """
                - CATEGORY_NAME: diffrn_detector
                  ATTRIBUTE_NAME_LIST:
                      - pdbx_frequency
                - CATEGORY_NAME: pdbx_serial_crystallography_measurement
                  ATTRIBUTE_NAME_LIST:
                      - diffrn_id
                      - pulse_energy
                      - pulse_duration
                      - xfel_pulse_repetition_rate
        """
        for catName in container.getObjNameList():
            cObj = container.getObj(catName)
            print("- CATEGORY_NAME: %s" % catName)
            print("  ATTRIBUTE_NAME_LIST:")
            for atName in cObj.getAttributeList():
                print("       - %s" % atName)
        return True

    def getIhmDevPathList(self):
        return self.__getIhmDevPathList(self.__getRepoTopPath('ihm_dev'))

    def __getIhmDevPathList(self, topRepoPath):
        """ Return the list of I/HM entries in the current repository.

            File name template is: PDBDEV_0000 0020_model_v1-0.cif.gz

            List is ordered in increasing PRDDEV numerical code.
        """
        pathList = []
        logger.debug("Searching path %r" % topRepoPath)
        try:
            sd = {}
            for root, dirs, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PDBDEV_") and name.endswith(".cif.gz") and len(name) <= 50:
                        pth = os.path.join(root, name)
                        sd[int(name[7:15])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing search in %r with %s" % (topRepoPath, str(e)))
        #
        return self.__applyFileLimit(pathList)



##
# File: HashableDict.py
#
##
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping


class HashableDict(Mapping):

    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __hash__(self):
        return hash(tuple(sorted(self._d.items())))

    def __repr__(self):
        return self._d.__repr__()


if __name__ == '__main__':
    hd = HashableDict()
    print(hd)

    td = HashableDict({'a': 1, 'b': 2})
    print(td)
    print(type(td))

    rd = HashableDict({'a': 1, 'b': 2, 'c': 3})
    l1 = [td, td, td]
    t1 = tuple(l1)
    l2 = [rd, rd, rd, rd]
    t2 = tuple(l2)
    ok = t1 == t2

    q = set(l1) == set(l2)
    print("l1-l2 %r" % q)

    q = set(l1) == set(l1)
    print("l1-l1 %r" % q)
    #
    kwD = HashableDict({'dictMapPath': 'pth string'})
    hd = HashableDict({'locator': 1, 'format': 'vrpt-xml-to-cif', 'kwargs': kwD})
    w1 = [hd, hd, hd]
    q = list(set(w1))
    print(q)
    #
